from datetime import datetime as dt


# prints a user friendly error message if the dob is not formatted properly
def check_dob_format(dob):
    try:
        if dob is not None:
            dt.strptime(dob, "%Y-%m-%d")
    except ValueError as error:
        error.args = (
            "The date `{0}` does not match the format YYYY-MM-DD".format(dob),
        )
        raise



from cytoolz.curried import curry
from collections import deque
from .base import FileSink
from govcf.variant_files import ensure_collection
from genomoncology.parse import DocType

import related


@related.immutable()
class DummyVariantRecord(object):
    chrom = related.StringField()
    pos = related.IntegerField()
    ref = related.StringField()
    alt = related.StringField()
    info = related.ChildField(dict, default=dict)

    CHR_FIELDS = ["chr", "c"]
    POS_FIELDS = ["pos", "s", "start", "hg19_pos", "hg37_pos", "hg38_pos"]
    REF_FIELDS = ["ref", "r"]
    ALT_FIELDS = ["alt", "a"]
    INFO_IGNORE = {"__type__", "build"}

    def __str__(self):
        return (
            "\t".join(
                [
                    self.chrom,
                    str(self.pos),
                    ".",
                    self.ref,
                    self.alt,
                    ".",
                    ".",
                    self.info_str,
                ]
            )
            + "\n"
        )

    @property
    def info_str(self):
        return ";".join(self.info_nvp())

    def info_nvp(self):
        for (name, value) in self.info.items():
            if name == "info":
                ivalues = value[0].items()
                for (iname, ivalue) in ivalues:
                    yield f"{iname}={ivalue}"

            elif name not in self.INFO_IGNORE:
                value_str = ",".join(map(str, value))
                yield f"{name}={value_str}"

    @classmethod
    def pop_first(cls, unit: dict, field_names: list):
        for field_name in field_names:
            value = unit.pop(field_name, None)
            if value is not None:
                return value

    @classmethod
    def create_from_unit(cls, unit: dict):
        unit = unit.copy()
        chrom = cls.pop_first(unit, cls.CHR_FIELDS)
        pos = cls.pop_first(unit, cls.POS_FIELDS)
        ref = cls.pop_first(unit, cls.REF_FIELDS)
        alt = cls.pop_first(unit, cls.ALT_FIELDS)

        info = {}
        for (name, value) in unit.items():
            info[name] = ensure_collection(value)

        return DummyVariantRecord(chrom, pos, ref, alt, info)


@curry
class VcfFileSink(FileSink):
    def __init__(self, filename):
        super().__init__(filename, insert_newlines=False)
        self.recent_records_seen = deque(maxlen=100)
        self.count = 0

    @classmethod
    def get_record_str(cls, unit) -> str:
        record = unit.get("__record__")
        if record is None:
            if DocType.HEADER.is_a(unit):
                record = DEFAULT_VCF_HEADER
            else:
                record = DummyVariantRecord.create_from_unit(unit)
        record = str(record)
        return record

    def convert(self, unit):
        self.count += 1
        record = self.get_record_str(unit)

        if record not in self.recent_records_seen:
            self.recent_records_seen.append(record)
            return record


DEFAULT_VCF_HEADER = """##fileformat=VCFv4.2
##contig=<ID=1,length=249250621>
##contig=<ID=2,length=243199373>
##contig=<ID=3,length=198022430>
##contig=<ID=4,length=191154276>
##contig=<ID=5,length=180915260>
##contig=<ID=6,length=171115067>
##contig=<ID=7,length=159138663>
##contig=<ID=8,length=146364022>
##contig=<ID=9,length=141213431>
##contig=<ID=10,length=135534747>
##contig=<ID=11,length=135006516>
##contig=<ID=12,length=133851895>
##contig=<ID=13,length=115169878>
##contig=<ID=14,length=107349540>
##contig=<ID=15,length=102531392>
##contig=<ID=16,length=90354753>
##contig=<ID=17,length=81195210>
##contig=<ID=18,length=78077248>
##contig=<ID=19,length=59128983>
##contig=<ID=20,length=63025520>
##contig=<ID=21,length=48129895>
##contig=<ID=22,length=51304566>
##contig=<ID=X,length=155270560>
##contig=<ID=Y,length=59373566>
##contig=<ID=MT,length=16569>
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
"""



from cytoolz.curried import assoc, compose, merge

from genomoncology import kms
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import (
    register,
    name_mapping,
    rename,
    split_value,
)


def set_chr_start(value):
    if value.get("build", "").endswith("38"):
        value = rename("#chr", "chr", value)
        value = rename("pos(1-based)", "start", value)
    else:
        value = rename("hg19_chr", "chr", value)
        value = rename("hg19_pos(1-based)", "start", value)

    return value


def get_largest_value(value):
    return max(value) if value else None


def additional_clean_up(value):
    genes = split_value(value.get("gene"))

    aa_changes = value.get("Transcript_var_VEST3__mstring")

    if len(genes) == 1 and len(aa_changes) > 1:
        genes = genes * len(aa_changes)

    alterations = [
        f"{gene} {aa_change}" for (gene, aa_change) in zip(genes, aa_changes)
    ]

    updates = dict(
        clinvar_clnsig=get_largest_value(value.get("clinvar_clnsig__mint")),
        gene=genes,
        alteration=alterations,
    )

    return merge(value, updates)


NAME_MAPPING = {
    # critical
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    "build": "build",
    "gene": "genename",
    "clinvar_clnsig__mint": "clinvar_clnsig",
    "VEST3_score__mfloat": "VEST3_score",
    "VEST3_rankscore__mfloat": "VEST3_rankscore",
    "Transcript_id_VEST3__mstring": "Transcript_id_VEST3",
    "Transcript_var_VEST3__mstring": "Transcript_var_VEST3",
    # baylor
    "1000Gp3_AC__int": "1000Gp3_AC",
    "1000Gp3_AF__float": "1000Gp3_AF",
    "1000Gp3_AFR_AC__int": "1000Gp3_AFR_AC",
    "1000Gp3_AFR_AF__float": "1000Gp3_AFR_AF",
    "1000Gp3_AMR_AC__int": "1000Gp3_AMR_AC",
    "1000Gp3_AMR_AF__float": "1000Gp3_AMR_AF",
    "1000Gp3_EAS_AC__int": "1000Gp3_EAS_AC",
    "1000Gp3_EAS_AF__float": "1000Gp3_EAS_AF",
    "1000Gp3_EUR_AC__int": "1000Gp3_EUR_AC",
    "1000Gp3_EUR_AF__float": "1000Gp3_EUR_AF",
    "1000Gp3_SAS_AC__int": "1000Gp3_SAS_AC",
    "1000Gp3_SAS_AF__float": "1000Gp3_SAS_AF",
    "cds_strand__string": "cds_strand",
    "MutationTaster_AAE__mstring": "MutationTaster_AAE",
    "MutationTaster_converted_rankscore__float": "MutationTaster_converted_rankscore",  # noqa E501
    "MutationTaster_model__mstring": "MutationTaster_model",
    "MutationTaster_pred__mstring": "MutationTaster_pred",
    "MutationTaster_score__mfloat": "MutationTaster_score",
    "Polyphen2_HDIV_pred__mstring": "Polyphen2_HDIV_pred",
    "Polyphen2_HDIV_rankscore__mfloat": "Polyphen2_HDIV_rankscore",
    "Polyphen2_HDIV_score__mfloat": "Polyphen2_HDIV_score",
    "Polyphen2_HVAR_pred__mstring": "Polyphen2_HVAR_pred",
    "Polyphen2_HVAR_rankscore__mfloat": "Polyphen2_HVAR_rankscore",
    "Polyphen2_HVAR_score__mfloat": "Polyphen2_HVAR_score",
    "SIFT_converted_rankscore__mfloat": "SIFT_converted_rankscore",
    "SIFT_pred__mstring": "SIFT_pred",
    "SIFT_score__mfloat": "SIFT_score",
    # not used (yet)
    "clinvar_golden_stars__mint": "clinvar_golden_stars",
    "PROVEAN_converted_rankscore__float": "PROVEAN_converted_rankscore",
    "PROVEAN_pred__mstring": "PROVEAN_pred",
    "PROVEAN_score__mfloat": "PROVEAN_score",
    "codonpos__int": "codonpos",
    "codon_degeneracy__string": "codon_degeneracy",
    "Ancestral_allele__string": "Ancestral_allele",
    "AltaiNeandertal__string": "AltaiNeandertal",
    "Denisova__string": "Denisova",
    "Ensembl_geneid__mstring": "Ensembl_geneid",
    "Ensembl_transcriptid__mstring": "Ensembl_transcriptid",
    "Ensembl_proteinid__mstring": "Ensembl_proteinid",
    "LRT_score__float": "LRT_score",
    "LRT_converted_rankscore__float": "LRT_converted_rankscore",
    "LRT_pred__string": "LRT_pred",
    "LRT_Omega__float": "LRT_Omega",
    "MutationAssessor_UniprotID__string": "MutationAssessor_UniprotID",
    "MutationAssessor_variant__string": "MutationAssessor_variant",
    "MutationAssessor_score__string": "MutationAssessor_score",
    "MutationAssessor_pred__string": "MutationAssessor_pred",
    "FATHMM_score__mfloat": "FATHMM_score",
    "FATHMM_converted_rankscore__float": "FATHMM_converted_rankscore",
    "FATHMM_pred__mstring": "FATHMM_pred",
    "MetaSVM_score__float": "MetaSVM_score",
    "MetaSVM_rankscore__float": "MetaSVM_rankscore",
    "MetaSVM_pred__string": "MetaSVM_pred",
    "MetaLR_score__float": "MetaLR_score",
    "MetaLR_rankscore__float": "MetaLR_rankscore",
    "MetaLR_pred__string": "MetaLR_pred",
    "Reliability_index__float": "Reliability_index",
    "M-CAP_score__float": "M-CAP_score",
    "M-CAP_rankscore__float": "M-CAP_rankscore",
    "M-CAP_pred__string": "M-CAP_pred",
    "MutPred_score__string": "MutPred_score",
    "MutPred_rankscore__string": "MutPred_rankscore",
    "MutPred_protID__string": "MutPred_protID",
    "MutPred_AAchange__string": "MutPred_AAchange",
    "MutPred_Top5features__string": "MutPred_Top5features",
    "fathmm-MKL_coding_score__float": "fathmm-MKL_coding_score",
    "fathmm-MKL_coding_rankscore__float": "fathmm-MKL_coding_rankscore",
    "fathmm-MKL_coding_pred__string": "fathmm-MKL_coding_pred",
    "fathmm-MKL_coding_group__string": "fathmm-MKL_coding_group",
    "Eigen_coding_or_noncoding__string": "Eigen_coding_or_noncoding",
    "Eigen-raw__float": "Eigen-raw",
    "Eigen-phred__float": "Eigen-phred",
    "Eigen-PC-raw__float": "Eigen-PC-raw",
    "Eigen-PC-phred__float": "Eigen-PC-phred",
    "Eigen-PC-raw_rankscore__float": "Eigen-PC-raw_rankscore",
    "GenoCanyon_score__float": "GenoCanyon_score",
    "GenoCanyon_score_rankscore__float": "GenoCanyon_score_rankscore",
    "integrated_fitCons_score__float": "integrated_fitCons_score",
    "integrated_confidence_value__float": "integrated_confidence_value",
    "GM12878_fitCons_score__float": "GM12878_fitCons_score",
    "GM12878_fitCons_rankscore__float": "GM12878_fitCons_score_rankscore",
    "GM12878_confidence_value__float": "GM12878_confidence_value",
    "H1-hESC_fitCons_score__float": "H1-hESC_fitCons_score",
    "H1-hESC_fitCons_rankscore__float": "H1-hESC_fitCons_score_rankscore",
    "H1-hESC_confidence_value__float": "H1-hESC_confidence_value",
    "HUVEC_fitCons_score__float": "HUVEC_fitCons_score",
    "HUVEC_fitCons_rankscore__float": "HUVEC_fitCons_score_rankscore",
    "HUVEC_confidence_value__float": "HUVEC_confidence_value",
    "GERP++_NR__float": "GERP++_NR",
    "GERP++_RS__float": "GERP++_RS",
    "GERP++_RS_rankscore__float": "GERP++_RS_rankscore",
    "phyloP100way_vertebrate__float": "phyloP100way_vertebrate",
    "phyloP100way_vertebrate_rankscore__float": "phyloP100way_vertebrate_rankscore",  # noqa E501
    "phyloP20way_mammalian__float": "phyloP20way_mammalian",
    "phyloP20way_mammalian_rankscore__float": "phyloP20way_mammalian_rankscore",  # noqa E501
    "phastCons100way_vertebrate__float": "phastCons100way_vertebrate",
    "phastCons100way_vertebrate_rankscore__float": "phastCons100way_vertebrate_rankscore",  # noqa E501
    "phastCons20way_mammalian__float": "phastCons20way_mammalian",
    "phastCons20way_mammalian_rankscore__float": "phastCons20way_mammalian_rankscore",  # noqa E501
    "SiPhy_29way_pi__mfloat": "SiPhy_29way_pi",
    "SiPhy_29way_logOdds__float": "SiPhy_29way_logOdds",
    "SiPhy_29way_logOdds_rankscore__float": "SiPhy_29way_logOdds_rankscore",
    "TWINSUK_AC__int": "TWINSUK_AC",
    "TWINSUK_AF__float": "TWINSUK_AF",
    "ALSPAC_AC__int": "ALSPAC_AC",
    "ALSPAC_AF__float": "ALSPAC_AF",
    "Interpro_domain__mstring": "Interpro_domain",
    "GTEx_V6p_gene__string": "GTEx_V6p_gene",
    "GTEx_V6p_tissue__string": "GTEx_V6p_tissue",
}

register(
    input_type=DocType.TSV,
    output_type=DocType.DBNSFP,
    transformer=compose(
        additional_clean_up,
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        lambda x: assoc(x, __TYPE__, DocType.DBNSFP.value),
        name_mapping(NAME_MAPPING, empty_values=(None, ".")),
        set_chr_start,
    ),
)

register(
    input_type=DocType.TSV,
    output_type=DocType.DBNSFP,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.GNOMAD.value)),
    is_header=True,
)



import asyncio
import os

import click
from click.types import Path
from cytoolz.curried import (
    filter,
    map,
    partition_all,
    concat,
    unique,
    curry,
    partitionby,
)
from specd import create_app

from genomoncology import kms
from genomoncology.parse import DocType, is_not_skipped_call
from genomoncology.pipeline import (
    converters,
    filters,
    run_pipeline,
    sinks,
    sources,
    transformers,
)
from gosdk import logger
from govcf import BEDFilter, iterate_vcf_files
from . import options
import uuid


# const

DEFAULT_SPECD_PATH = os.path.join(
    os.path.dirname(__file__), "../../gosdk/specs/"
)

VERSION = "0.8.19"
COPYRIGHT = "Copyright: GenomOncology, LLC"


def print_version(ctx, _, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"GO CLI: Version {VERSION}")
    ctx.exit()


@click.group(chain=True, invoke_without_command=True)
@click.argument("input", type=str)
@click.argument("output", type=str)
@click.option(
    "--dir",
    "dir",
    help="Use this to specify the directory "
    "where your input and output "
    "files are located.",
)
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
)
@options.common_options
@options.pass_state
@click.pass_context
def gocli(ctx, state, input, output, dir):
    logger.setup_logging(quiet=state.quiet, verbose=state.debug)
    state.set_start()
    state.loop = asyncio.get_event_loop()
    ctx.call_on_close(state.finalize)


# register: genomoncology.pipeline.runner.run_pipeline
gocli.resultcallback()(run_pipeline)


@gocli.command()
@click.option(
    "--bed",
    "-b",
    help="BED file for filtering VCF records.",
    type=Path(exists=True),
)
@click.option(
    "--padding",
    "-p",
    help="Padding added to BED regions.",
    type=int,
    default=0,
)
@click.option(
    "--keep_ref_unknown_calls",
    "-k",
    "keep_ref_unknown_calls",
    help="Include homozygous reference and unknown GT variants as calls",
    is_flag=True,
)
@options.glob_option
@options.build_option
@options.pass_state
def from_vcf(state, bed, padding, keep_ref_unknown_calls):
    """Parses VCF into variant calls objects."""
    bed_filter = BEDFilter(bed, padding=padding) if bed else None
    return [
        sources.CollectFilePathsSource(glob=state.glob),
        curry(iterate_vcf_files)(
            include_vaf=True,
            bed_filter=bed_filter,
            build=state.build,
            keep_ref_unknown_calls=keep_ref_unknown_calls,
        ),
    ]


@gocli.command()
@click.option(
    "--genes",
    "-g",
    "gene_column",
    help="Specify if bed file has gene column",
    is_flag=True,
)
@click.option("--skip_comment", type=bool, default=False)
@click.option("--comment_char", type=str, default="#")
@options.build_option
@options.pass_state
def from_bed(state, gene_column, skip_comment, comment_char):
    """ Read records from a BED file source. """
    columns = ["chromosome", "start", "end"]
    if gene_column:
        columns.append("gene")
    return [
        sources.DelimitedFileSource(
            columns=columns,
            delimiter="\t",
            build=state.build,
            skip_comment=skip_comment,
            comment_char=comment_char,
        )
    ]


@gocli.command()
@click.option(
    "--column",
    "-c",
    "columns",
    help="Specify columns of TSV source.",
    type=str,
    multiple=True,
)
@click.option("--skip_comment", type=bool, default=False)
@click.option("--comment_char", type=str, default="#")
@options.build_option
@options.pass_state
def from_tsv(state, columns, skip_comment, comment_char):
    """Parses TSV into name-value pair objects."""
    return [
        sources.DelimitedFileSource(
            columns=columns,
            delimiter="\t",
            build=state.build,
            skip_comment=skip_comment,
            comment_char=comment_char,
        )
    ]


@gocli.command()
@click.argument("aggregate_key", type=str)
@options.build_option
@options.pass_state
def aggregate_tsv(state, aggregate_key):
    """ Rolls up TSV records using a specified key. """
    return [
        sources.AggregatedFileSource(
            aggregate_key=aggregate_key, delimiter="\t"
        )
    ]


@gocli.command()
@click.argument("aggregate_key", type=str)
@options.build_option
@options.pass_state
def aggregate_csv(state, aggregate_key):
    """ Rolls up CSV records using a specified key. """
    return [
        sources.AggregatedFileSource(
            aggregate_key=aggregate_key, delimiter=","
        )
    ]


@gocli.command()
@click.option(
    "--column",
    "-c",
    "columns",
    help="Specify columns of TSV source.",
    type=str,
    multiple=True,
)
@options.build_option
@options.pass_state
def from_csv(state, columns):
    """Parses CSV into name-value pair objects."""
    return [
        sources.DelimitedFileSource(
            columns=columns, delimiter=",", build=state.build
        )
    ]


@gocli.command()
def from_excel():
    """Parses Excel into name-value pair objects."""
    # todo: merge with parse using --format
    return [sources.ExcelSource]


@gocli.command()
@click.option(
    "--column",
    "-c",
    "columns",
    help="Specify columns of TSV source.",
    type=str,
    multiple=True,
)
@options.build_option
@options.pass_state
def from_maf(state, columns):
    """ Read records from a MAF file source. """
    return [sources.MAFSource(columns=columns, build=state.build)]


@gocli.command()
@click.option(
    "--fields",
    "-f",
    help="Add field to standard annotation set.",
    type=str,
    multiple=True,
)
@click.option(
    "--async",
    "-a",
    "run_async",
    help="Run in asynchronous mode.",
    is_flag=True,
)
@click.option(
    "--delete-if-exists",
    "-d",
    "delete_if_exists",
    help="Delete if annotations exist.",
    is_flag=True,
)
@options.build_option
@options.pass_state
def annotate_calls(state, fields, run_async, delete_if_exists):
    """Get annotations for stream of calls."""
    if run_async:
        create_function = kms.create_async_processor
        sdk_function = kms.annotations.async_annotate
    else:
        create_function = kms.create_sync_processor
        sdk_function = kms.annotations.sync_annotate

    return [
        filter(is_not_skipped_call),
        partition_all(50),  # annotate calls is GET, hardcoded batch size
        create_function(
            state,
            sdk_function,
            fields=fields,
            delete_if_exists=delete_if_exists,
            build=state.build,
        ),
        concat,
    ]


@gocli.command()
@options.build_option
@options.pass_state
def annotate_genes(state):
    """Get gene objects by stream of names."""
    return [
        sources.TextFileSource,
        partition_all(state.batch_size),
        kms.create_sync_processor(
            state, kms.genes.sync_boundaries, build=state.build
        ),
        concat,
    ]


@gocli.command()
@click.argument("glom_path", type=str)
@click.argument("cmp", type=str)
@click.argument("value", type=str)
def filter_in(glom_path, cmp, value):
    """Filter in objects that match comparison."""
    return filter(filters.glom_include(glom_path, cmp, value))


@gocli.command()
@click.argument("glom_path", type=str)
@click.argument("cmp", type=str)
@click.argument("value", type=str)
def filter_out(glom_path, cmp, value):
    """Filter out objects that match comparison."""
    return filter(filters.glom_exclude(glom_path, cmp, value))


@gocli.command()
@click.argument("glom_path", type=str)
@click.argument("cmp", type=str)
@click.argument("value", type=str)
def retain(glom_path, cmp, value):
    """Always keep objects that match comparison."""
    return map(filters.mark_retain(glom_path, cmp, value))


@gocli.command()
@click.option(
    "--column",
    "-c",
    help="Specify column for TSV display.",
    type=str,
    multiple=True,
)
@click.option("--include-header", "-h", is_flag=True)
def to_tsv(column, include_header):
    """Render objects to TSV file format."""
    return [sinks.TsvFileSink(columns=column, include_header=include_header)]


@gocli.command()
def to_vcf():
    """Render calls to VCF file format."""
    return [sinks.VcfFileSink()]


@gocli.command()
def to_excel():
    """Render objects to Excel file format."""
    return [sinks.ExcelSink()]


@gocli.command()
def to_python():
    """Render python objects without JSON transform."""
    return sinks.TextFileSink


@gocli.command()
def to_pretty():
    """Render indented, syntax highlighted JSON."""
    return map(converters.to_pretty_json_str)


@gocli.command()
@click.option(
    "--diseases",
    "-d",
    help="Disease(s) for match eligibility.",
    type=str,
    multiple=True,
)
@click.option("--gender", help="Gender for match eligibility.", type=str)
@click.option(
    "--dob",
    help="Date of birth for match eligibility. Format: YYYY-MM-DD",
    type=str,
)
@options.pass_state
def match_trials(state, diseases, gender, dob):
    """Match trials by variants and disease."""
    from . import utils

    utils.check_dob_format(dob)
    return [
        partition_all(99_999_999),
        kms.create_sync_processor(
            state,
            kms.trials.match_trials,
            diseases=diseases,
            gender=gender,
            dob=dob,
        ),
        concat,
    ]


@gocli.command()
@click.option(
    "--diseases",
    "-d",
    help="Disease(s) for match eligibility.",
    type=str,
    multiple=True,
)
@options.pass_state
def match_therapies(state, diseases):
    """Match therapies by variants and disease."""
    return [
        partition_all(99_999_999),
        kms.create_sync_processor(
            state, kms.therapies.match_therapies, diseases=diseases
        ),
        concat,
        filter(converters.non_null),
    ]


@gocli.command()
@click.option(
    "--diseases",
    "-d",
    help="Disease(s) for match eligibility.",
    type=str,
    multiple=True,
)
@options.pass_state
def match_contents(state, diseases):
    """Match contents by variants and disease."""
    return [
        partition_all(99_999_999),
        kms.create_sync_processor(
            state, kms.contents.match_contents, diseases=diseases
        ),
        concat,
        filter(converters.non_null),
    ]


@gocli.command()
@options.pipeline_option
@options.run_id_option
@options.pass_state
@click.argument("output_type", type=str)
def transform(state, output_type):
    """Transform input stream to output type."""
    return [
        filter(is_not_skipped_call),
        transformers.create_transformer(output_type=output_type, state=state),
        filter(None),
    ]


@gocli.command()
@click.argument("field", type=str)
def extract(field):
    """Extract a specific field from object stream."""
    return map(transformers.get_in_field(field))


@gocli.command()
def flatten():
    """Flatten list of lists into a list of strings."""
    return [filter(converters.non_null), concat]


@gocli.command()
def distinct():
    """Remove duplicates from a list of strings."""
    return unique


@gocli.command()
@click.option("--name", "-n", help="Spec Name", type=str, default=None)
@click.option("--target", "-t", help="Spec Targets", multiple=True)
@options.quiet_option
@options.pass_state
def swagger(state, name, target):
    """Launches local Swagger UI webserver."""
    cwd = os.getcwd()
    os.chdir(DEFAULT_SPECD_PATH)
    create_app(
        include_swagger=True, host=state.host, name=name, target=target
    ).run()
    os.chdir(cwd)
    return map(print("Exited Swagger"))


@gocli.command()
@click.option(
    "--workers",
    help="Number of load workers (default: 10)",
    type=int,
    default=10,
)
@options.size_option
@options.build_option
@options.pass_state
def load_warehouse(state, workers):
    """Loads variants to warehouse."""
    return [
        filter(is_not_skipped_call),
        partition_all(state.batch_size),
        sinks.LoadWarehouseVariantsSink(state=state, num_workers=workers),
        concat,
    ]


@gocli.command()
@click.option(
    "--workers",
    help="Number of load workers (default: 10)",
    type=int,
    default=10,
)
@options.size_option
@options.build_option
@options.pass_state
def load_features(state, workers):
    """Loads variants to warehouse."""
    return [
        partition_all(state.batch_size),
        sinks.LoadWarehouseFeaturesSink(state=state, num_workers=workers),
        concat,
    ]


@gocli.command()
@click.argument("data_set", type=str)
@click.argument("data_set_version", type=str)
@click.option(
    "--workers",
    help="Number of load workers (default: 10)",
    type=int,
    default=10,
)
@options.size_option
@options.build_option
@options.pass_state
def load_annotations(state, data_set, data_set_version, workers):
    """Loads variant objects into annotations core."""
    return [
        filter(is_not_skipped_call),
        transformers.create_transformer(output_type=data_set, state=state),
        filter(None),
        partition_all(state.batch_size),
        sinks.LoadAnnotationSink(
            state=state,
            data_set=data_set,
            data_set_version=data_set_version,
            num_workers=workers,
        ),
        concat,
        sinks.JsonlFileSink,
    ]


@gocli.command()
@options.size_option
@click.option(
    "--hgvs", "recalculate_hgvs", help="Full HGVS refresh.", is_flag=True
)
@click.option("--bed", "-b", help="Bed file", type=str, default=None)
@options.pass_state
def refresh_annotations(state, recalculate_hgvs, bed):
    """Rebuilds annotations merged core."""
    source = sources.BedFileSource(bed_file=bed) if bed else sources.NullSource
    return [
        source,
        kms.create_sync_processor(
            state,
            kms.annotations.sync_refresh,
            recalculate_hgvs=recalculate_hgvs,
            batch_size=state.batch_size,
        ),
    ]


@gocli.command()
@options.pass_state
def region_search(state):
    """Searches for Transcript in region specified by bed file"""
    return [
        filter(DocType.HEADER.is_not),
        kms.create_sync_processor(state, kms.transcripts.process_transcripts),
        sinks.JsonlFileSink,
    ]


@gocli.command()
@click.option(
    "--arg", "-a", "args", help="Argument: name=value", type=str, multiple=True
)
@click.argument("source_name", type=str)
@options.pass_state
def from_source(state, source_name, args):
    """Parse file into name-value pair objects."""
    # TODO: good enough for uniprot and mitomap, but refactor later
    source = sources.get_one_off_source(source_name)
    assert source, f"Source not found: {source_name}"
    kwargs = {}
    if args:
        kwargs = dict([a.split("=") for a in args])
    return [source(build=state.build, **kwargs)]


@gocli.command()
@click.option(
    "--arg", "-a", "args", help="Argument: name=value", type=str, multiple=True
)
@click.argument("function_name", type=str)
def invoke(function_name, args):
    """Invoke an external function."""
    import importlib

    module_name, function_name = function_name.rsplit(".", 1)
    module = importlib.import_module(module_name)

    f = getattr(module, function_name)
    f = curry(f)

    if args:
        kwargs = dict([a.split("=") for a in args])
        f = f(**kwargs)

    return map(f)


@gocli.command()
@click.argument("bed_file", type=str)
@click.argument("flag_name", type=str)
def add_flag(bed_file, flag_name):
    """Add flag based on BED file value/presence."""
    bed_filter = BEDFilter(bed_file)
    return map(transformers.add_flag(bed_filter, flag_name))


@gocli.command()
def version():
    print(f"GO CLI: Version {VERSION} [{COPYRIGHT}]")
    return [sources.NullSource]


@gocli.command()
def rollup():
    """Roll up variants by record."""
    return [
        partitionby(lambda r: r.get("__record__", uuid.uuid4())),
        map(transformers.add_rollup),
        concat,
    ]



from .base import BaseInfoMapper, register

AD = "AD"
DP = "DP"


class USCInfoMapper(BaseInfoMapper):
    @classmethod
    def keys(cls):
        return {DP, AD}

    def do_mapping(self, call_info, allele_position, alt_allele):
        total_depth = call_info.get(DP)
        allele_depths = call_info.get(AD)
        if allele_depths and len(allele_depths) > allele_position:
            ref_depth = allele_depths[0]
            alt_depth = allele_depths[allele_position]
        else:
            alt_depth, ref_depth = None, None  # pragma: no cover

        return total_depth, alt_depth, ref_depth


register(USCInfoMapper)



import urllib3

from .state import State

from . import cli

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


__author__ = """Ian Maurer"""
__email__ = "ian@genomoncology.com"
__version__ = cli.commands.VERSION

__uri__ = "http://pypi.org/project/gocli"
__copyright__ = "Copyright (c) 2018 genomoncology.com"
__description__ = "gocli: GenomOncology Command Line Interface"
__doc__ = __description__ + " <" + __uri__ + ">"
__title__ = "gocli"


__all__ = ("State", "cli")



from .variant_files import iterate_vcf_files, iterate_vcf_calls
from .bed_filter import BEDFilter


__all__ = ("iterate_vcf_files", "iterate_vcf_calls", "BEDFilter")

__author__ = """Ian Maurer"""
__email__ = "ian@genomoncology.com"
__version__ = "0.8.0"

__uri__ = "http://pypi.org/project/govcf"
__copyright__ = "Copyright (c) 2018 genomoncology.com"
__description__ = "govcf: GenomOncology VCF Call Generator"
__doc__ = __description__ + " <" + __uri__ + ">"
__title__ = "govcf"



from .factory import create_sync_processor, create_async_processor
from . import annotations, contents, genes, therapies, transcripts, trials

__all__ = (
    "create_sync_processor",
    "create_async_processor",
    "annotations",
    "contents",
    "genes",
    "therapies",
    "transcripts",
    "trials",
)



from cytoolz.curried import curry, assoc, filter

from genomoncology.parse import ensures, DocType, __TYPE__, __CHILD__
from .base import LazyFileSource
import csv


def do_split(delimiter, s):
    s = s.strip()
    if delimiter == ",":
        # reference: https://stackoverflow.com/a/35822856
        val = next(csv.reader([s]))
    else:
        val = s.split(delimiter)
    return val


@curry
def str_to_dict(columns, meta, delimiter, skip_comment, comment_char, s):
    if skip_comment:
        if s.startswith(comment_char):
            return None
    # note: warning, doesn't handle escaping for CSVs yet.
    d = dict(zip(columns, do_split(delimiter, s)))
    d.update(meta)
    return d


@curry
class DelimitedFileSource(LazyFileSource):
    def __init__(
        self,
        filename,
        columns,
        delimiter="\t",
        skip_comment=False,
        comment_char="#",
        **meta
    ):
        self._mapper = None
        self.columns = ensures.ensure_collection(columns)
        self.delimiter = delimiter
        self.skip_comment = skip_comment
        self.comment_char = comment_char
        self.meta = meta

        if __TYPE__ not in meta:
            self.meta = assoc(self.meta, __TYPE__, DocType.TSV.value)

        super().__init__(filename)

    def __iter__(self):
        # noinspection PyUnresolvedReferences
        iterator = super(DelimitedFileSource.func, self).__iter__()

        if not self.columns:
            data_line = next(iterator).strip()
            self.columns = do_split(self.delimiter, data_line)

        yield self.create_header()

        # noinspection PyTypeChecker
        yield from filter(lambda x: x is not None, map(self.mapper, iterator))

    def create_header(self):
        return {
            __TYPE__: DocType.HEADER.value,
            __CHILD__: self.meta.get(__TYPE__),
            "columns": self.columns,
            "meta": self.meta,
            "file_path": self.name,
        }

    @property
    def mapper(self):
        if self._mapper is None:
            self._mapper = str_to_dict(
                self.columns,
                self.meta,
                self.delimiter,
                self.skip_comment,
                self.comment_char,
            )
        return self._mapper



from genomoncology import kms
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    # hgvs
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    "build": "build",
    # info
    "CNT__int": "info.CNT",
    "gene": "info.GENE",
    "CDS__string": "info.CDS",
    "AA__string": "info.AA",
    "ID__string": "info.ID",
    "tissues__mstring": "info.TISSUES",
    "resistance_mutation__mstring": "info.RESISTANCE_MUTATION",
    "tissue_frequency__mstring": "info.TISSUE_FREQ",
}

register(
    input_type=DocType.CALL,
    output_type=DocType.COSMIC,
    transformer=compose(
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        lambda x: assoc(x, __TYPE__, DocType.COSMIC.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.CALL,
    output_type=DocType.COSMIC,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.COSMIC.value)),
    is_header=True,
)



from genomoncology import kms
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    # hgvs
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    "build": "build",
    # info
    "DSS__int": "info.DSS",
    "WSSF__float": "info.WSSF",
    "VSSF__float": "info.VSSF",
    "WME__float": "info.WME",
    "VME__float": "info.VME",
    "LWME__float": "info.LWME",
    "LVME__float": "info.LVME",
    "LWNNS__float": "info.LWNNS",
    "LVNNS__float": "info.LVNNS",
    "LWHSF__float": "info.LWHSF",
    "LVHSF__float": "info.LVHSF",
}

register(
    input_type=DocType.CALL,
    output_type=DocType.SPLICING,
    transformer=compose(
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        lambda x: assoc(x, __TYPE__, DocType.SPLICING.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.CALL,
    output_type=DocType.SPLICING,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.SPLICING.value)),
    is_header=True,
)



from .mitomap import MitomapFileSource
from .uniprot import UniprotFileSource

source_name_map = {"mitomap": MitomapFileSource, "uniprot": UniprotFileSource}

__all__ = ("MitomapFileSource", "UniprotFileSource")



from .base import BaseInfoMapper, register
from .utils import int_or_zero

DP = "DP"
DP4 = "DP4"


class PheoInfoMapper(BaseInfoMapper):
    @classmethod
    def keys(cls):
        return {DP, DP4}

    def do_mapping(self, call_info, allele_position, alt_allele):
        total_depth = call_info.get(DP)

        dp4 = call_info.get(DP4, None)
        alt_depth, ref_depth = None, None
        if dp4 and (len(dp4) == 4):
            ref_depth = int_or_zero(dp4[0]) + int_or_zero(dp4[1])
            alt_depth = int_or_zero(dp4[2]) + int_or_zero(dp4[3])

        return total_depth, alt_depth, ref_depth


register(PheoInfoMapper, -1)



def int_or_none(value):
    try:
        return int(value)
    except (TypeError, ValueError):
        pass


def int_or_zero(value):
    try:
        return int(value)
    except (TypeError, ValueError):  # pragma: no cover
        return 0



from cytoolz.curried import curry, dissoc

from gosdk import logger


@curry
def process_transcripts(record, sdk=None):
    record = dissoc(record, "__type__", "build")
    logger.get_logger().debug("get_transcripts", **record)

    results = sdk.call_with_retry(
        sdk.transcripts.get_transcripts, **dissoc(record, "gene")
    )["results"]

    return {**record, "transcripts": results}



# common.py - models that are shared across versions.

from specd.sdk import BaseModel


class AnnotationList(BaseModel):

    # Set of keys that are NOT annotations. New top-level keys must be added.
    META_KEYS = {"errors", "is_created", "additionalProperties", "normalized"}

    def get_annotation(self, key):
        obj_dict = getattr(self, key, None)
        if obj_dict:
            return self.instantiate(self.definitions.Annotation, obj_dict)

    def iter_annotations(self):
        for key in self:
            if key not in AnnotationList.META_KEYS:
                yield self.get_annotation(key)



from genomoncology.parse.doctypes import DocType, __TYPE__, __CHILD__
from .registry import register, get_transformer


class VariantMapper(object):
    """
    Variant is a Call with additional meta information such as file_path,
    build, pipeline, run_id.

    Also, all of the 'info' fields are inserted to to the top-level of the
    document with wild-card typing format (e.g. DP__mint, etc.)
    """

    def __init__(self, **meta):
        self.meta = {k: v for (k, v) in meta.items() if v is not None}
        self.types = {}

    def resolve_info_key(self, k):
        return f"{k}__{self.types.get(k, 'string')}"

    def __call__(self, doc):
        if DocType.HEADER.is_a(doc):
            self.types = doc.get("types", {})
            self.meta["file_path"] = doc.get("file_path", None)

            # todo: replace copy with cytoolz
            header = doc.copy()
            header[__CHILD__] = doc.get(__CHILD__, "").replace(
                "CALL", "VARIANT"
            )
            return header

        elif DocType.VARIANT.is_a(doc):
            return doc

        elif DocType.TSV.is_a(doc) or DocType.ANNOTATED_TSV.is_a(doc):
            variant = doc.copy()
            variant.pop("annotations", None)
            variant[__TYPE__] = "VARIANT"
            variant.update(self.meta)
            return variant

        elif DocType.CALL.is_a(doc) or DocType.ANNOTATED_CALL.is_a(doc):
            # todo: replace copy with cytoolz
            # todo: replace with registered function, remove this elif
            variant = doc.copy()
            variant.pop("annotations", None)
            variant[__TYPE__] = doc.get(__TYPE__).replace("CALL", "VARIANT")
            variant.update(self.meta)
            info = variant.pop("info", {})
            info = {self.resolve_info_key(k): v for (k, v) in info.items()}
            variant.update(info)
            return variant

        else:
            tx = get_transformer(DocType.VARIANT, doc.get(__TYPE__))
            return tx(doc)


register(transformer=VariantMapper, output_type=DocType.VARIANT)


class FeatureMapper(object):
    r"""
    Feature Fields:
        * chr
        * start
        * end

    File Fields:
        * build
        * run_id
        * file_path
        * pipeline

    Info Fields:
        * \d+           => __int
        * \d+\.\d+      => __float
        * anything else => __string
        * w/ comma      => __m(int,float,string)
    """

    CORE = {"chr", "start", "end", "build", "run_id", "file_path", "pipeline"}

    def __init__(self, **meta):
        self.meta = {k: v for (k, v) in meta.items() if v is not None}
        self.meta[__TYPE__] = "FEATURE"
        self.types = {}
        self.is_multi = set()

    @staticmethod
    def get_val_type(value):
        is_multi = False
        if "," in value and len(value) > 1:
            is_multi = True
            value = value.split(",")[0]

        try:
            int(value)
            val_type = "int"
        except ValueError:
            try:
                float(value)
                val_type = "float"
            except ValueError:
                val_type = "string"

        val_type = "m" + val_type if is_multi else val_type
        return is_multi, val_type

    def transform_key_value(self, key, value):
        if key not in self.types:
            is_multi, val_type = self.get_val_type(value)
            if is_multi:
                self.is_multi.add(key)
            self.types[key] = val_type

        if value and (key in self.is_multi):
            value = value.split(",")

        return f"{key}__{self.types[key]}", value

    def __call__(self, doc: dict):  # pragma: no mccabe
        if DocType.HEADER.is_a(doc):
            self.meta["file_path"] = doc.get("file_path", None)
            header = doc.copy()
            header[__CHILD__] = "FEATURE"
            return header

        elif DocType.FEATURE.is_a(doc):
            return doc

        elif DocType.TSV.is_a(doc):
            feature = self.meta.copy()
            for (key, value) in doc.items():
                if key in self.CORE:
                    feature[key] = value
                elif key == "stop":
                    feature["end"] = value
                elif key != "__type__":
                    new_key, new_value = self.transform_key_value(key, value)
                    feature[new_key] = new_value

            return feature

        else:
            tx = get_transformer(DocType.VARIANT, doc.get(__TYPE__))
            return tx(doc)


register(transformer=FeatureMapper, output_type=DocType.FEATURE)



from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping


NAME_MAPPING = {
    "gene": "gene",
    # info
    "disease_name__mstring": "disease name",
    "gene_mim__mstring": "gene mim",
    "disease_mim__mstring": "disease mim",
    "DDD_category__mstring": "DDD category",
    "allelic_requirement__mstring": "allelic requirement",
    "mutation_consequence__mstring": "mutation consequence",
    "phenotypes__mstring": "phenotypes",
    "organ_specificity_list__mstring": "organ specificity list",
    "pmids__mstring": "pmids",
    "panel__mstring": "panel",
    "prev_symbols__mstring": "prev symbols",
    "hgnc_id__mstring": "hgnc id",
    "gene_disease_pair_entry_date__mstring": "gene disease pair entry date",
}

register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.BAYLOR_DECIPHER,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __TYPE__, DocType.BAYLOR_DECIPHER.value),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.AGGREGATE,
    output_type=DocType.BAYLOR_DECIPHER,
    transformer=compose(
        lambda x: assoc(x, "is_gene_annotation", True),
        lambda x: assoc(x, __CHILD__, DocType.BAYLOR_DECIPHER.value),
    ),
    is_header=True,
)



from .base import BaseInfoMapper, register

TIR = "TIR"
DP = "DP"


class StrelkaInfoMapper(BaseInfoMapper):
    @classmethod
    def keys(cls):
        return {TIR, DP, "AU", "CU", "GU", "TU"}

    def do_mapping(self, call_info, allele_position, alt_allele):
        total_depth = call_info.get(DP)

        if TIR in call_info:
            alt_depth = call_info.get(TIR, [None])[0]
        else:
            alt_depth = call_info.get("%sU" % alt_allele, [None])[0]

        ref_depth = (
            (total_depth - alt_depth)
            if None not in (total_depth, alt_depth)
            else None
        )

        return total_depth, alt_depth, ref_depth


register(StrelkaInfoMapper)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('./src/genomoncology/README.md') as readme_file:
    readme = readme_file.read()

history = ""

tests_require = [
    "pytest-flake8 >= 1.0.2",
    "black >= 18.5b1",
    "ipython >= 6.4.0",
    "pytest >= 3.6.0",
    "pytest-cov >= 2.5.1",
    "pytest-mccabe >= 0.1",
    "white >= 0.1.2",
    "pytest-socket >= 0.2.0",
    "pytest-asyncio >= 0.8.0",
    "aioconsole >= 0.1.8",
    "addict >= 2.1.3",
    "requests >= 2.4",
]

setup(
    name="gocli",
    version='0.8.19',
    author="Ian Maurer",
    author_email='ian@genomoncology.com',

    packages=[
        "genomoncology",
        "genomoncology.kms",
        "genomoncology.cli",
        "genomoncology.parse",
        "genomoncology.pipeline",
        "genomoncology.pipeline.sinks",
        "genomoncology.pipeline.sources",
        "genomoncology.pipeline.sources.one_off_sources",
        "genomoncology.pipeline.transformers",
        "genomoncology.pipeline.transformers.tx",
        "gosdk",
        "govcf",
        "govcf.calculate_vaf",

    ],
    package_dir={
        '': 'src'
    },

    package_data={
        '': ["*.yaml", "*.bed", "*.txt", "*.tsv", "*.csv"]
    },

    include_package_data=True,

    tests_require=tests_require,

    install_requires=[
        "specd >= 0.8.1",
        "backoff >= 1.5.0",
        "click == 6.7",
        "structlog >= 18.1.0",
        "colorama >= 0.3.9",
        "pysam >= 0.14.1",
        "ujson >= 1.3.5",
        "intervaltree == 2.1.0",
        "glom >= 18.1.1",
        "cytoolz >= 0.9.0.1",
        "openpyxl < 2.5.10",
        "pygments >= 2.2.0",
        "jsonschema[format] >= 2.6.0",
        "flask == 1.0.2",
    ],

    setup_requires=[
        'pytest-runner',
    ],

    license="Proprietary",
    keywords='Bioinformatics HGVS VCF Clinical Trials Genomics',

    description="gocli",
    long_description="%s\n\n%s" % (readme, history),

    entry_points={
        'console_scripts': [
            'gocli=genomoncology.main:main',
        ],
    },

    classifiers=[
        'License :: Other/Proprietary License',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)



from genomoncology import kms
from cytoolz.curried import assoc, compose
from genomoncology.parse.doctypes import DocType, __CHILD__, __TYPE__
from genomoncology.pipeline.transformers import register, name_mapping
from genomoncology.pipeline.transformers.functions import filter_keys


ORIGIN_MAPPING = {
    "0": "unknown",
    "1": "germline",
    "2": "somatic",
    "4": "inherited",
    "8": "paternal",
    "16": "maternal",
    "32": "de-novo",
    "64": "biparental",
    "128": "uniparental",
    "256": "not-tested",
    "512": "tested-inconclusive",
    "1073741824": "other",
}

NAME_MAPPING = {
    # hgvs
    "chr": "chr",
    "start": "start",
    "ref": "ref",
    "alt": "alt",
    # info
    "CLNSIG__mstring": "info.CLNSIG",
    "ID__string": "info.RS",
    "CLNREVSTAT__mstring": "info.CLNREVSTAT",
    "CLNVI__mstring": "info.CLNVI",
    "ORIGIN__mstring": "info.ORIGIN",
}


def clean_origin(value):
    value = [ORIGIN_MAPPING.get(value[int(i)]) for i in range(0, len(value))]
    return value


register(
    input_type=DocType.CALL,
    output_type=DocType.CLINVAR,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.CLINVAR.value),
        filter_keys(
            {
                "CLNSIG__mstring",
                "ID__string",
                "CLNREVSTAT__mstring",
                "CLNVI__mstring",
                "ORIGIN__mstring",
                "hgvs_g",
            }
        ),
        lambda x: assoc(
            x, "ORIGIN__mstring", clean_origin(x.get("ORIGIN__mstring", ""))
        ),
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        name_mapping(NAME_MAPPING),
    ),
)

register(
    input_type=DocType.CALL,
    output_type=DocType.CLINVAR,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.CLINVAR.value)),
    is_header=True,
)



from cytoolz.curried import assoc, compose

from genomoncology import kms
from genomoncology.parse.doctypes import DocType, __TYPE__, __CHILD__
from genomoncology.pipeline.transformers.functions import filter_keys
from genomoncology.pipeline.transformers.registry import register
from genomoncology.pipeline.transformers.mapper_classes import VariantMapper


SSR_MAPPING = {
    0: "unspecified",
    1: "Paralog",
    2: "byEST",
    4: "oldAlign",
    8: "Para_EST",
    16: "1kg_failed",
    1024: "other",
    3: "Paralog,byEST",
    5: "oldAlign,Paralog",
    6: "oldAlign,byEST",
    9: "Para_EST,Paralog",
    10: "Para_EST,byEST",
    12: "Para_EST,oldAlign",
    11: "Para_EST,Paralog,byEST",
    13: "Para_EST,Paralog,oldAlign",
    14: "Para_EST,byEST,oldAlign",
    17: "Paralogkg_failed,Paralog",
    18: "Paralogkg_failed,byEST",
    20: "Paralogkg_failed,oldAlign",
    24: "Paralogkg_failed,Para_EST",
    19: "Paralogkg_failed,Paralog,byEST",
    21: "Paralogkg_failed,Paralog,oldAlign",
    25: "Paralogkg_failed,Paralog,Para_EST",
    22: "Paralogkg_failed,byEST,oldAlign",
    26: "Paralogkg_failed,byEST,Para_EST",
    28: "Paralogkg_failed,oldAlign,Para_EST",
    23: "Paralogkg_failed,Paralog,byEST,oldAlign",
    27: "Paralogkg_failed,Paralog,byEST,Para_EST",
    30: "Paralogkg_failed,byEST,oldAlign,Para_EST",
    31: "Paralogkg_failed,Paralog,byEST,oldAlign,Para_EST",
    1025: "other,Paralog",
    1026: "other,byEST",
    1028: "other,oldAlign",
    1032: "other,Para_EST",
    1040: "other,Paralogkg_failed",
    1027: "other,Paralog,byEST",
    1029: "other,Paralog,oldAlign",
    1033: "other,Paralog,Para_EST",
    1041: "other,Paralog,Paralogkg_failed",
    1030: "other,byEST,oldAlign",
    1034: "other,byEST,Para_EST",
    1042: "other,byEST,Paralogkg_failed",
    1036: "other,oldAlign,Para_EST",
    1044: "other,oldAlign,Paralogkg_failed",
    1048: "other,Para_EST,Paralogkg_failed",
    1031: "other,Paralog,byEST,oldAlign",
    1035: "other,Paralog,byEST,Para_EST",
    1043: "other,Paralog,byEST,Paralogkg_failed",
    1037: "other,Paralog,oldAlign,Para_EST",
    1045: "other,Paralog,oldAlign,Paralogkg_failed",
    1049: "other,Paralog,Para_EST,Paralogkg_failed",
    1038: "other,byEST,oldAlign,Para_EST",
    1046: "other,byEST,oldAlign,Paralogkg_failed",
    1050: "other,byEST,Para_EST,Paralogkg_failed",
    1052: "other,oldAlign,Para_EST,Paralogkg_failed",
    1039: "other,Paralog,byEST,oldAlign,Para_EST",
    1047: "other,Paralog,byEST,oldAlign,Paralogkg_failed",
    1053: "other,Paralog,oldAlign,Para_EST,Paralogkg_failed",
    1054: "other,byEST,oldAlign,Para_EST,Paralogkg_failed",
    1055: "other,Paralog,byEST,oldAlign,Para_EST,Paralogkg_failed",
}


def clean_ssr(value):
    if type(value) == int:
        value = SSR_MAPPING.get(value)
    return value


register(
    input_type=DocType.CALL,
    output_type=DocType.DBSNP,
    transformer=compose(
        lambda x: assoc(x, __TYPE__, DocType.DBSNP.value),
        filter_keys({"VLD__string", "SSR__string", "RS__string", "hgvs_g"}),
        # clean SSR: replace origin int code with string
        lambda x: assoc(x, "SSR__string", clean_ssr(x.get("SSR__string", ""))),
        lambda x: assoc(x, "hgvs_g", kms.annotations.to_csra(x)),
        # convert Call to Variant
        VariantMapper(),
    ),
    is_header=False,
)

register(
    input_type=DocType.CALL,
    output_type=DocType.DBSNP,
    transformer=compose(lambda x: assoc(x, __CHILD__, DocType.DBSNP.value)),
    is_header=True,
)



import os
import tarfile
from fnmatch import fnmatch
from glob import iglob

from cytoolz.curried import curry

from genomoncology.parse.ensures import ensure_collection
from .base import Source
from .const import TGZ_PATH_DELIMITER, CLOUD_RE_PATTERN


@curry
class CollectFilePathsSource(Source):
    def __init__(self, path, glob=None, include_tar=False):
        self.path = path
        self.glob = glob
        self.include_tar = include_tar

    def __iter__(self):
        yield from collect_files(self.path, self.glob, self.include_tar)


def is_cloud_path(path):
    return CLOUD_RE_PATTERN.match(path) is not None


def find_files(paths, glob=None, include_tar=False):
    for path in ensure_collection(paths or ["."]):

        if path.startswith("~") and os.path.exists(os.path.expanduser(path)):
            path = os.path.expanduser(path)  # pragma: no cover

        # todo: handle "buckets", currently assumes single files only.
        if is_cloud_path(path):
            yield path

        elif os.path.isfile(path):
            yield path

        else:
            glob = ensure_collection(glob) or [None]
            yield from find_normal_files(glob, path)
            if include_tar:
                yield from find_tarred_files(glob, path)


def find_tarred_files(glob, path):
    for g in ["*.tgz", "*.tar.gz", "*.tar"]:
        pattern = ["**", g] if g else ["**"]
        pattern = os.path.join(path, *pattern)

        for file_path in iglob(pattern, recursive=True):
            file_path = os.path.realpath(file_path)
            tgz = tarfile.open(file_path)
            for member_path in tgz.getnames():
                member_name = os.path.basename(member_path)
                for g in ensure_collection(glob):
                    if fnmatch(member_name, g):
                        components = ["tgz", file_path, member_path]
                        yield TGZ_PATH_DELIMITER.join(components)


def find_normal_files(glob, path):
    for g in glob:
        pattern = ["**", g] if g else ["**"]
        pattern = os.path.join(path, *pattern)

        for file_path in iglob(pattern, recursive=True):
            yield os.path.realpath(file_path)


def collect_files(paths, glob=None, include_tar=False, **_):
    """ collect files finds all of the paths and sorts them alphabetically. """
    yield from find_files(paths, glob=glob, include_tar=include_tar)
